/*	xmms ao output plugin
 *	Copyright (C) 2003 Levi Ramsey
 *
 *	This program is Free Software and licensed under the GNU GPL
 *
 *	Description: Outputs audio from xmms to libao2.  This enables automatic
 *		detection of sound system to use (OSS, ALSA, esd, aRts, etc.)
 *
 *	Based on Galex Yen's esd output plugin
 *
 *	Contains code (C) 1998-1999 Mikael Alm, Olle Hallnas, Thomas Nilsson,
 *		and 4Front Technologies
 */

#include "aoout.h"
#include <ao/ao.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <pthread.h>

#include <xmms/xmmsctrl.h>
#include <xmms/configfile.h>

static aoout_format output_format, input_format;
static gboolean swapsign, going, buffer;
static ao_device *device;
static guint64 written, output_time_offset, output_bytes;
static int paused=0;
static AFormat sfmt;
static pthread_t bufplayer;
static int buflen, wcur=0, rcur=0, flush=-1;
static char* buf;
static int blk_size;
static int lvol=-1, rvol=-1;
static gboolean fileout=FALSE;
static guint8 *peaks8;
static gint16 *peaks16;
static int numpeaks, peaki;

extern gfloat lastgain[2];

static aoout_format aoout_set_format(AFormat fmt, int rate, int nch);
static void *aoout_play_buffer(void *arg);
static void aoout_write_buffer(void *ptr, int len);
static int aoout_buffer_used(void);
static void aoout_play_audio(void *ptr, int length);
static ao_device *aoout_open_file(int did);
static void aoout_do_amplify_audio(void *ptr, int length, gfloat l, gfloat r);
static void aoout_amplify_audio(void *ptr, int length);

extern gint ctrlsocket_get_session_id(void);

void aoout_get_volume(int *l, int *r) {
	*l=lvol;
	*r=rvol;
}

void aoout_set_volume(int l, int r) {
	lvol=l;
	rvol=r;
/*	fprintf(stderr, "Left: %i\tRight:%i\n", l, r);	*/
}

int aoout_open(AFormat fmt, int rate, int nch) {
	int did;
	int i;

	buffer=ao_cfg.buffering;

	ao_initialize();
	output_format=aoout_set_format(fmt, rate, nch);
	input_format=output_format;
	
	device=NULL;
	if (ao_cfg.fileout) {
		device=aoout_open_file(-1);
	} else {
		did=ao_default_driver_id();
		if (did==-1) {
			ao_shutdown();
			return 0;
		}

		device=ao_open_live(did, &(output_format.ao), NULL);
		if (!device) {
			if (errno==AO_ENOTLIVE) {
				/*	fprintf(stderr, "xmmsao: File output not supported\n");	*/
				device=aoout_open_file(did);
			}
		}
	}
	
	if (!device) {
		ao_shutdown();
		return 0;
	}

	paused=0;
	going=TRUE;

	written=output_time_offset=output_bytes=0;
	flush=-1;

	numpeaks=1800/ao_cfg.bufchunk+1;
	peaki=0;
	peaks8=calloc(numpeaks, sizeof(guint8));
	if (!peaks8) {
		going=FALSE;
		ao_close(device);
		ao_shutdown();
		return 0;
	}
	for (i=0; i<numpeaks; i++) {
		peaks8[i]=0;
	}
	peaks16=calloc(numpeaks, sizeof(gint16));
	if (!peaks16) {
		going=FALSE;
		free(peaks8);
		ao_close(device);
		ao_shutdown();
		return 0;
	}
	for (i=0; i<numpeaks; i++) {
		peaks16[i]=0;
	}

	if (buffer) {
		blk_size=ao_cfg.bufchunk*output_format.bytespersecond/1000;
		buflen=ao_cfg.bufsize*output_format.bytespersecond/1000;
		buf=calloc(buflen, sizeof(char));
		if (!buf) {
			going=FALSE;
			ao_close(device);
			ao_shutdown();
			return 0;
		}
		pthread_create(&bufplayer, NULL, aoout_play_buffer, NULL);
	}

	lvol=ao_cfg.lvol;
	rvol=ao_cfg.rvol;

	return 1;
}

aoout_format aoout_set_format(AFormat fmt, int rate, int nch) {
	ao_sample_format output_format;
	aoout_format retstruct;
	gboolean swap;

	sfmt=fmt;
	swap=FALSE;
	if (fmt==FMT_S8 || fmt==FMT_U8) {
		output_format.bits=8;
		if (fmt==FMT_S8) {
			swapsign=TRUE;
		}
	} else {
		output_format.bits=16;
		if (fmt==FMT_U16_LE || fmt==FMT_U16_BE || fmt==FMT_U16_NE) {
			swap=TRUE;
		}
		switch (fmt) {
			case FMT_U16_LE:
			case FMT_S16_LE:
				output_format.byte_format=AO_FMT_LITTLE;
				break;
			case FMT_U16_BE:
			case FMT_S16_BE:
				output_format.byte_format=AO_FMT_BIG;
				break;
			case FMT_U16_NE:
			case FMT_S16_NE:
				output_format.byte_format=AO_FMT_NATIVE;
				break;
		}
	}
	output_format.rate=rate;
	output_format.channels=nch;

	retstruct.bytespersecond=output_format.bits/8*rate*nch;
	retstruct.ao=output_format;
	retstruct.swap=swapsign=swap;

	return retstruct;
}

void *aoout_play_buffer(void *arg) {
	int length, cnt;

	while (going) {
		if (!paused) {
			length=((blk_size<aoout_buffer_used()) ? (blk_size) : (aoout_buffer_used()));
			if (length<4) {
				continue;
			}

			length-=(length%4);

			while (length>0) {
				cnt=((length<(buflen-rcur)) ? (length) : (buflen-rcur));
				aoout_play_audio(buf+rcur, cnt);
				rcur=(rcur+cnt)%buflen;
				length-=cnt;
			}
		} else {
			usleep(5000);
		}

		if (flush!=-1) {
			output_time_offset=flush;
			written=(guint64)(flush/10)*(guint64)(output_format.bytespersecond/100);
			rcur=wcur=output_bytes=0;
			flush=-1;
		}
	}

	pthread_exit(NULL);
}

void aoout_write_buffer(void *ptr, int len) {
	int cnt, off=0;

	while (len>0) {
		cnt=((len<(buflen-wcur)) ? (len) : (buflen-wcur));
		memcpy((char *)buf+wcur, (char *)ptr+off, cnt);
		wcur=(wcur+cnt)%buflen;
		len-=cnt;
		off+=cnt;
	}
}	

int aoout_buffer_used(void) {
	if (wcur>=rcur) {
		return wcur-rcur;
	}
	return buflen-(rcur-wcur);
}

void aoout_write(void *ptr, int length) {
	written+=length;

	if (buffer) {
		aoout_write_buffer(ptr, length);
	} else {
		aoout_play_audio(ptr, length);
	}
}

void aoout_play_audio(void *ptr, int length) {
	AFormat newfmt;
	int newfreq, newchan;
	EffectPlugin *ep;

	if (!ptr || length<=0 || paused) {
		return;
	}

	if (ao_cfg.prevolume) {
		aoout_amplify_audio(ptr, length);
	}

	newfmt=sfmt;
	newfreq=output_format.ao.rate;
	newchan=output_format.ao.channels;

	ep=get_current_effect_plugin();

	if (((!fileout) || (fileout && ao_cfg.effectfile)) && effects_enabled() && ep) {
		if (ep->query_format) {
			ep->query_format(&newfmt, &newfreq, &newchan);
		}
		if (newfmt==sfmt && newfreq==output_format.ao.rate && newchan==output_format.ao.channels && ep->mod_samples) {
			ep->mod_samples(&ptr, length, sfmt, output_format.ao.rate, output_format.ao.channels);
		}
	}

	if (!ao_cfg.prevolume) {
		aoout_amplify_audio(ptr, length);
	}

		ao_play(device, ptr, length);
	output_bytes+=length;
}

void aoout_close(void) {
	ConfigFile *cfgfile;

	going=FALSE;

	if (buffer) {
		pthread_join(bufplayer, NULL);
		free(buf);
		wcur=0;
		rcur=0;
	}

	free(peaks8);
	free(peaks16);

	if (!((ao_cfg.lvol==lvol) || (ao_cfg.rvol==rvol))) {
		cfgfile=xmms_cfg_open_default_file();
		xmms_cfg_write_int(cfgfile, "aoout", "lvol", lvol);
		xmms_cfg_write_int(cfgfile, "aoout", "rvol", rvol);
		xmms_cfg_write_default_file(cfgfile);
		xmms_cfg_free(cfgfile);

		ao_cfg.lvol=lvol;
		ao_cfg.rvol=rvol;
	}
	
	ao_close(device);

	ao_shutdown();

	fileout=FALSE;
}

void aoout_flush(int time) {
	if (buffer) {
		flush=time;
		while (flush!=-1) {
			usleep(5000);
		}
	} else {
		output_time_offset=time;
		written=(time/10)*(output_format.bytespersecond/100);
		output_bytes=0;
	}
}

void aoout_pause(short p) {
	paused=p;
}

int aoout_free(void) {
	if (buffer) {
		if (rcur>wcur) {
			return (rcur-wcur)-1;
		}
		return (buflen-(wcur-rcur))-1;
	}
	
	if (paused)
		return 0;

	return 1000000;
}

int aoout_playing(void) {
	if (!going) {
		return FALSE;
	}
	if (!aoout_buffer_used()) {
		return FALSE;
	}

	return TRUE;
}

int aoout_get_output_time(void) {
	int output_time;
	if (!output_format.bytespersecond || !going)
		return 0;
	
	output_time=output_time_offset+(int)(output_bytes*1000/output_format.bytespersecond);

	return output_time;
}

int aoout_get_written_time(void) {
	if (!output_format.bytespersecond || !going)
		return 0;

	return (int)((written*1000)/output_format.bytespersecond);
}

/*	Adapted from the disk_writer xmms plugin	*/
ao_device *aoout_open_file(did) {
	char *filename, *title, *tmp;
	int pos;

	char *type;
	ao_device *d;

	pos=xmms_remote_get_playlist_pos(ctrlsocket_get_session_id());
	title=xmms_remote_get_playlist_file(ctrlsocket_get_session_id(), pos);
	if (title!=NULL && (tmp=strrchr(title, '.')) != NULL)
		*tmp=NULL;

	if (title==NULL || strlen(g_basename(title))==0) {
		g_free(title);

		title=xmms_remote_get_playlist_title(ctrlsocket_get_session_id(), pos);
		while (title!=NULL && (tmp=strchr(title, '/')) != NULL)
			*tmp='-';

		if (title==NULL || strlen(g_basename(title))==0) {
			g_free(title);

			title=g_strdup_printf("xmms-%d", pos);
		}
	}

	if (did==-1) {
		type=ao_cfg.filetype;
		did=ao_driver_id(type);
		
		if (did==-1) {
			g_free(title);
			fprintf(stderr, "xmmsao: %s is not a supported output file type\n", type);
			return NULL;
		}
	} else {
		type=(ao_driver_info(did))->short_name;
	}

	filename=g_strdup_printf("%s/%s.%s", ao_cfg.directory, g_basename(title), g_basename(type));
	g_free(title);

	d=ao_open_file(did, filename, ao_cfg.fileoverwrite, &(output_format.ao), NULL);
	if (!d) {
		if (errno==AO_EFILEEXISTS) {
			fprintf(stderr, "xmmsao: %s already exists... not writing\n", filename);
		}
	}

	g_free(filename);

	fileout=TRUE;
	
	return d;
}

void aoout_amplify_audio(void *ptr, int length) {
	gfloat left=0.0;
	gfloat right=0.0;

	left=((gfloat)(lvol))/100.0;
	right=((gfloat)(rvol))/100.0;
	
	aoout_do_amplify_audio(ptr, length, left, right);
}

void aoout_do_amplify_audio(void *ptr, int length, gfloat left, gfloat right) {
	static gfloat cur;
	gint out;
	guint8 *ptr8;
	gint16 *ptr16;
	gboolean stereo=(output_format.ao.channels==2);
	int len;

	len=length;

	cur=((stereo) ? (left) : ((left+right)/2.0));

	if (output_format.swap)
		return;

	if (output_format.ao.bits==8) {
		ptr8=(gint8 *)ptr;

		while (length--) {
			out=(gint)((gfloat)(*ptr8)*cur);
			if (out>255) {
				out=255;
			} else if (out<0) {
				out=0;
			}
			*(ptr8++)=out;
			
			if (stereo) {
				cur=((cur==left) ? (right) : (left));
			}
		}
	} else {
		ptr16=(gint16 *)ptr;
		length/=2;

		while (length--) {
			out=*ptr16;
			switch (output_format.ao.byte_format) {
				case AO_FMT_NATIVE:
#ifdef WORDS_BIGENDIAN
				case AO_FMT_BIG:
#else
				case AO_FMT_LITTLE:
#endif
					break;
				/*	Need to swap bytes	*/
				default:
					out=(out<<8)+(out>>8);
			}

			out=(gint)((gfloat)(out)*cur);
			if (out>32767) {
				out=32767;
			} else if (out<-32768) {
				out=-32768;
			}
			*ptr16=out;

			if (stereo) {
				cur=((cur==left) ? (right) : (left));
			}
			ptr16++;
		}
	}
}
